@@ -22,15 +22,52 @@
 
 from __future__ import annotations
 
-from dataclasses import dataclass
+from dataclasses import asdict, dataclass
+from typing import Any
 
-__all__ = ("ServerInfo", "ClusterInfo")
+__all__ = ("ServerInfo", "ClusterInfo", "BaseInfo", "BrainInfo")
+
+
+class BaseInfo:
+    uid: int
+    _info_classes: dict[int, type[BaseInfo]] = {}
+    _info_class_id: int
+
+    def __init_subclass__(cls) -> None:
+        if not hasattr(cls, "_info_class_id"):
+            raise AttributeError(f"{cls} is missing _info_class_id.")
+        if used_by := BaseInfo._info_classes.get(cls._info_class_id):
+            raise ValueError(
+                f"_info_class_id {cls._info_class_id} is already used by "
+                f"{used_by}"
+            )
+        BaseInfo._info_classes[cls._info_class_id] = cls
+
+    def asdict(self) -> dict[str, Any]:
+        """Convert this info class to a dictionary."""
+
+        dct = asdict(self)
+        dct["_info_class_id"] = self._info_class_id
+        return dct
+
+    @staticmethod
+    def fromdict(data: dict[str, Any]) -> BaseInfo:
+        """Convert a dictionary back into its info class."""
+
+        cls = BaseInfo._info_classes[data.pop("_info_class_id")]
+        return cls(**data)
+
+    def __int__(self) -> int:
+        """The IPC UID of the client this info class represents."""
+
+        return self.uid
 
 
 @dataclass
-class ServerInfo:
+class ServerInfo(BaseInfo):
     """A representation of a :class:`~server.Server`."""
 
+    _info_class_id = 0
     uid: int
     """The ipc uid of the server."""
     cluster_uids: list[int]
@@ -38,9 +75,10 @@ class ServerInfo:
 
 
 @dataclass
-class ClusterInfo:
+class ClusterInfo(BaseInfo):
     """A representation of a :class:`~cluster.Cluster`."""
 
+    _info_class_id = 1
     uid: int
     """The ipc uid of the cluster."""
     server_uid: int
@@ -69,3 +107,12 @@ class ClusterInfo:
         Assumes that all the shard ids of a cluster are adjacent."""
 
         return shard_id // shards_per_cluster
+
+
+@dataclass
+class BrainInfo(BaseInfo):
+    """A representation of a :class:`~brain.Brain`."""
+
+    _info_class_id = 2
+    uid: int
+    """The ipc uid of the brain."""
